'''
@desc    The broker connector class.
@author  arthuchaut <arthuchaut@gmail.com>
@version 0.1.0
@date    2019-09-20
@note    0.1.0 (2019-09-20): Writed the first drafts.
'''

import asyncio
from asyncio.events import AbstractEventLoop

import aiormq
from aiormq import Connection, Channel
from aiormq.exceptions import ProbableAuthenticationError


class Connector:
    '''Class allowing to create a connection to the broker.

    Attributes:
        _DEFAULT_PROTOCOL (str): The broker communication protocol.
        _DEFAULT_HOST (str): The broker hostname or ip address.
        _DEFAULT_VHOST (str): The connection virtual host.
        _DEFAULT_PORT (int): The broker listening port.
        _DEFAULT_USER (str): The username to connect.
        _DEFAULT_PASSWORD (str): The password to connect.

        _protocol (str): The broker communication protocol. 
        _host (str): The broker hostname or ip address.
        _vhost (str): The connection virtual host.
        _port (int): The broker listening port.
        _user (str): The username to connect.
        _password (str): The password to connect.
        _connection (Connection): The broker established connection.
        event_loop (AbstractEventLoop): The asyncio event loop.
            Try to get the running event loop, else create a new instance.
        _url (str): The built url for establishing the broker connection.
        connection (Connection): The broker established connection.
            Create a connection if isn't exists.
        channel (Channel): The connection channel.
    '''

    _DEFAULT_PROTOCOL: str = 'amqp'
    _DEFAULT_HOST: str = 'localhost'
    _DEFAULT_VHOST: str = '/'
    _DEFAULT_PORT: int = 5672
    _DEFAULT_USER: str = 'guest'
    _DEFAULT_PASSWORD: str = 'guest'

    def __init__(
        self,
        *,
        protocol: str = _DEFAULT_PROTOCOL,
        host: str = _DEFAULT_HOST,
        vhost: str = _DEFAULT_VHOST,
        port: int = _DEFAULT_PORT,
        user: str = _DEFAULT_USER,
        password: str = _DEFAULT_PASSWORD,
    ):
        '''The Connector initializer.

        Args:
            _protocol (str): The broker communication protocol. 
                Default to _DEFAULT_PROTOCOL.
            _host (str): The broker hostname or ip address.
                Default to _DEFAULT_HOST.
            _vhost (str): The connection virtual host.
                Default to _DEFAULT_VHOST.
            _port (int): The broker listening port.
                Default to _DEFAULT_PORT.
            _user (str): The username to connect.
                Default to _DEFAULT_USER.
            _password (str): The password to connect.
                Default to _DEFAULT_PASSWORD.
            _connection (Connection): The broker established connection.
            _url (str): The built url for establishing the broker connection.
            connection (Connection): The broker established connection.
                Create a connection if isn't exists.
            channel (Channel): The connection channel.
        '''

        self._protocol: str = protocol
        self._host: str = host
        self._vhost: str = Connector._DEFAULT_VHOST + vhost
        self._port: int = port
        self._user: str = user
        self._password: str = password
        self._connection: Connection = None

    @property
    def event_loop(self) -> AbstractEventLoop:
        '''Try to retrieve the current event loop if exists.
        
        Retreive the running event loop. If it doesn't exists yet,
        then create a new event loop.

        Returns:
            AbstractEventLoop: The asyncio event loop.
        '''

        loop: AbstractEventLoop = None

        try:
            loop = asyncio.get_running_loop()
        except Exception as e:
            loop = asyncio.get_event_loop()

        return loop

    @property
    def _url(self) -> str:
        '''The url builder property.
        Formate an url for establishing the broker connection.

        Returns:
            str: The built url.
        '''

        return (
            f'{self._protocol}://{self._user}:{self._password}'
            f'@{self._host}{self._vhost}'
        )

    @property
    async def connection(self) -> Connection:
        '''The broker connection.
        Create a connection if it isn't created yet.

        Raises:
            ConnectorError: If the connection couldn't be established.

        Returns:
            Connection: The broker connection.
        '''

        if self._connection is None:
            try:
                self._connection = await aiormq.connect(
                    self._url, loop=self.event_loop
                )
            except Exception as e:
                raise ConnectorError(e)

        return self._connection

    @property
    async def channel(self) -> Channel:
        '''The connection channel.

        Returns:
            Channel: The connection channel.
        '''

        return await (await self.connection).channel()


class ConnectorError(Exception):
    '''The ConnectorError exception class.
    '''



'''
@desc    The RPC class definition.
@author  arthuchaut <arthuchaut@gmail.com>
@version 0.1.0
@date    2019-09-21
@note    0.1.0 (2019-09-21): Writed the first drafts.
'''

from typing import Dict, Callable, List
from uuid import uuid4
from asyncio import Task, Future

import aiormq

from .utils import Props
from .message import IncomingMessage, SendingMessage
from .queue import Queue
from .timeout import Timeout


class RPC:
    '''The RPC class declaration.

    Allows to call the broker and waiting for its response back.

    Attributes:
        symbios (Symbios): A Symbios instance.
        _future (Future): The future response back.
        _ack (bool): Allows to know if the response was sent.
        _reply_queue (str): The queue to tell the broker where to reply.
    '''

    def __init__(self, symbios: object):
        '''The RPC initializer.

        Args:
            symbios (Symbios): The symbios instance.
        '''

        self.symbios: object = symbios
        self._futures: Dict[str, Future] = {}
        self._reply_queue: str = None

    async def _on_reply(
        self, symbios: object, message: IncomingMessage
    ) -> None:
        '''The task that will be called on the broker response.

        Valorize the self.response attribute with the received message.

        Args:
            message (IncomingMessage): The message from the broker.
        '''

        try:
            future, countdown = self._futures.pop(message.props.correlation_id)

            countdown.stop()
            future.set_result(message)
        except Exception as e:
            raise RPCError(f'Expected a correlation_id from the server.')

    async def call(
        self, message: SendingMessage, *, routing_key: str, timeout: int = None
    ) -> IncomingMessage:
        '''The calling procedure.

        Emit a message to the broker and wait for its response back.
        
        Args:
            message (SendingMessage): The message to send.
            routing_key (str): The routing key to emit the message.
            timeout (int): The time limite (in second) for the timeout.
                Raised a TimeoutElapsed exception if the countdown has
                expired. Default to None.

        Returns:
            IncomingMessage: The broker response back.
        '''

        cid: str = str(uuid4())
        countdown: Timeout = Timeout(self.symbios.event_loop)
        future: Future = self.symbios.event_loop.create_future()

        self._futures[cid] = (future, countdown)

        queue: Queue = Queue(cid, auto_delete=True)

        await self.symbios.declare_queue(queue)
        await self.symbios.listen(self._on_reply, queue=queue, no_ack=True)

        await self.symbios.emit(
            message,
            routing_key=routing_key,
            props=Props(correlation_id=cid, reply_to=queue.queue),
        )

        if timeout:
            await countdown.start(timeout)

        return await future

    def multi_calls(self, calls: List[call]) -> List[Task]:
        '''Process multi asynchronous RPC.

        Args:
            calls (List[RPC.call]): The list of RPC.call.

        Returns:
            List[Task]: The asyncio.Task list created from calls.
        '''

        task_queue: List[Task] = []

        for call in calls:
            task_queue.append(self.symbios.event_loop.create_task(call))

        return task_queue


class RPCError(Exception):
    '''The RPCError exception class.
    '''




import pamqp

from pamqp.specification import Basic
from pamqp.specification import Exchange
from pamqp.specification import Queue
from aiormq.types import ConfirmationFrameType


class EmitACK:
    '''The producer frame confirmation class.
    '''

    def __init__(self, confirmation: ConfirmationFrameType):
        self.confirmation: ConfirmationFrameType = confirmation


class ListenACK:
    '''The consumer frame confirmation class.
    '''

    def __init__(self, confirmation: Basic.ConsumeOk):
        self.confirmation: Basic.ConsumeOk = confirmation


class ExchangeACK:
    '''The exchange declaraiton frame confirmation.
    '''

    def __init__(self, confirmation: Exchange.DeclareOk):
        self.confirmation: Exchange.DeclareOk = confirmation


class QueueACK:
    '''The queue declaration frame confirmation.
    '''

    def __init__(self, confirmation: Queue.DeclareOk):
        self.confirmation: Queue.DeclareOk = confirmation



'''
@desc    The Symbios middleware manager classes.
@author  arthuchaut <arthuchaut@gmail.com>
@version 0.1.0
@date    2019-09-22
@note    0.1.0 (2019-09-22): Writed the first drafts.
'''

from typing import Union, List
from enum import Enum
from abc import ABC, abstractmethod

from .message import IncomingMessage, SendingMessage


class Event(Enum):
    '''The Event enumeration declaration.

    Allows to define the ownership of the event of a middleware.

    Attributes:
        ON_EMIT (int): Refer to the emitter.
        ON_LISTEN (int): Refer to the listener.
    '''

    ON_EMIT: int = 0
    ON_LISTEN: int = 1


class MiddlewareABC(ABC):
    '''The MiddlewareABC declaration.

    Abstract class for defining middleware.

    Attributes:
        event (int): The event associated with middleware.
    '''

    def __init__(self, event: int):
        '''The MiddlewareABC initializer.

        Args:
            event (int): The event associated with middleware.
        '''

        self.event: int = event

    @abstractmethod
    async def execute(
        self, symbios: object, message: Union[IncomingMessage, SendingMessage]
    ) -> None:
        '''Abstract method that contains the middleware processes.

        Args:
            symbios (Symbios): The Symbios instance.
            message (Union[IncomingMessage, SendingMessage]): 
                The received or sending message.
        '''

        pass


class MiddlewareLibrary:
    '''The MiddlewareLibrary class declaration.

    Allows to register the defined middlewares.

    Attributes:
        _library (Dict[int, List[MiddlewareABC]]): The library that 
            contains all declared middlewares.
    '''

    def __init__(self):
        '''The MiddlewareLibrary initializer.
        '''

        self._library: Dict[int, List[MiddlewareABC]] = {
            Event.ON_EMIT: [],
            Event.ON_LISTEN: [],
        }

    def append(self, midd: MiddlewareABC) -> None:
        '''Register a middleware to the library.

        Args:
            midd (MiddlewareABC): The middleware instance to register.

        Raises:
            MiddlewareLibraryError: If the midd argument is not a subclass
                of MiddlewareABC.
        '''

        if not isinstance(midd, MiddlewareABC):
            raise MiddlewareLibraryError(
                f'Expected an instance of MiddlewareABC, {type(midd)} given.'
            )

        self._library[midd.event].append(midd)

    async def run_until_end(
        self, symbios: object, message: SendingMessage, event: int
    ) -> None:
        '''Call all defined middlewares associated the event specified.

        Args:
            symbios (Symbios): The Symbios instance.
            message (Union[IncomingMessage, SendingMessage]): The 
                listener/emiter message.
        '''

        for midd in self._library[event]:
            await midd.execute(symbios, message)


class MiddlewareLibraryError(Exception):
    '''The MiddlewareLibraryError exception class.
    '''

    def __init__(self, message: str):
        '''The MiddlewareLibraryError initializer.

        Args:
            message (str): The message to raise.
        '''

        super().__init__(message)



'''
@desc    The producer class for emit requests to the broker.
@author  arthuchaut <arthuchaut@gmail.com>
@version 0.1.0
@date    2019-09-20
@note    0.1.0 (2019-09-20): Writed the first drafts.
'''

from typing import Awaitable, Any, Dict

from .utils import Props, Channel
from .message import SendingMessage
from .exchange import Exchange
from .middleware import MiddlewareLibrary, Event
from .confirmation import EmitACK, ExchangeACK


class Producer:
    '''The Producer class declaration.

    Allows to emit a message to the borker.

    Attributes:
        _CONTENT_TYPES (Dict[Any, str]): The association of types 
            with their content-type.
        symbios (Symbios): The Symbios instance.
        exchange (str): the exchange name to bind with the exchange type.
        exchange_type (str): The exhange type.
        routing_key (str): The message routing key.
        props (Props): The message properties that contain the header.
        mandatory (bool): Tell if the message is important or not.
        immediate (bool): Bypass the exchange type rules and send
            the message to the receiver directly if True.
        _midd_library (MiddlewareLibrary): The Symbios middleware library.        
    '''

    _CONTENT_TYPES: Dict[Any, str] = {
        str: 'text/plain',
        bool: 'text/plain',
        int: 'text/plain',
        float: 'text/plain',
        dict: 'application/json',
    }

    def __init__(
        self,
        *,
        symbios: object,
        exchange: Exchange = Exchange(),
        routing_key: str = None,
        props: Props = Props(),
        mandatory: bool = False,
        immediate: bool = False,
        midd_library: MiddlewareLibrary,
    ):
        '''The Producer initializer.

        Args:
            symbios (Symbios): The Symbios instance.
            exchange (Exchange): the exchange name to bind with the 
                exchange type. Default to Exchange().
            routing_key (str): The message routing key. Default to None.
            props (Props): The message properties that contain the header.
                Default to an empty instance of Props.
            mandatory (bool): Tell if the message is important or not.
                Default to False.
            immediate (bool): Bypass the exchange type rules and send
                the message to the receiver directly if True.
                Default to False.
            _midd_library (MiddlewareLibrary): The Symbios middleware library.    
        '''

        self.symbios: object = symbios
        self.exchange: Exchange = exchange
        self.routing_key: str = routing_key
        self.props: Props = props
        self.mandatory: bool = mandatory
        self.immediate: bool = immediate
        self.declare_ok: object = None
        self.produce_ok: object = None
        self._midd_library: MiddlewareLibrary = midd_library

    async def emit(self, message: SendingMessage) -> EmitACK:
        '''Emit a message to the broker.

        Important: Make sure that the destination queue(s) are correctly 
        declared by a Consumer before sending a message.
        Call all associated middlewares before emit the message.

        Args:
            message (SendingMessage): The message to send.

        Raises:
            ProducerError: If the type of exchange requires a routing_key.

        Returns:
            EmitACK: The producer confirmation.
        '''

        chann: Channel = await self.symbios.channel

        if self.exchange.exchange != '' and self.exchange.exchange is not None:
            declare_ok = await chann.exchange_declare(**self.exchange.__dict__)
            self.declare_ok = ExchangeACK(declare_ok)

        if not self.exchange.exchange_type in [
            Exchange.FANOUT,
            Exchange.HEADERS,
        ] and (self.routing_key == '' or self.routing_key is None):
            raise ProducerError(
                (
                    f'Exchange type {self.exchange.exchange_type} '
                    f'require a routing_key.'
                )
            )

        if not self.props.content_type:
            self.props.content_type = self._determine_content_type(message)

        if not self._midd_library is None:
            await self._midd_library.run_until_end(
                self.symbios, message, Event.ON_EMIT
            )

        produce_ok = await chann.basic_publish(
            message.serialized,
            routing_key=self.routing_key,
            exchange=self.exchange.exchange,
            properties=self.props,
            immediate=self.immediate,
            mandatory=self.mandatory,
        )

        self.produce_ok = EmitACK(produce_ok)

        return self.produce_ok

    def _determine_content_type(self, message: SendingMessage) -> str:
        '''Try to determine the content-type via the message type.

        Args:
            message (SendingMessage): The message to send.

        Returns:
            str: The content_type retreived or ''.
        '''

        content_type: str = ''

        try:
            content_type = Producer._CONTENT_TYPES[type(message.body)]
        except KeyError as e:
            pass

        return content_type


class ProducerError(Exception):
    '''The ProducerError Exception class.
    '''

    def __init__(self, message: str):
        super().__init__(message)



'''
@desc    The timeout class.
@author  arthuchaut <arthuchaut@gmail.com>
@version 0.1.0
@date    2019-09-23
@note    0.1.0 (2019-09-23): Writed the first drafts.
'''

import asyncio
from asyncio import AbstractEventLoop, Task, CancelledError


class Timeout:
    '''The Timeout class declaration.

    Allows to define a delay and raise a TimeoutElapsed exception
    if the countdown expire.

    Attributes:
        _loop (AbstractEventLoop): The asyncio event loop.
        _task (Task): A coroutine that process the countdown.
    '''

    def __init__(
        self, loop: AbstractEventLoop = asyncio.get_event_loop()
    ) -> None:
        '''The Timeout initializer.

        Args:
            _loop (AbstractEventLoop): The asyncio event loop.
                Default to a new event loop.
        '''

        self._loop: AbstractEventLoop = loop
        self._task: Task = None

    async def start(self, limit: int) -> None:
        '''Start the countdown.

        Create a task that start the countdown.
        Raise an exception if the task has ended.
        Return None if the task is cancelled.

        Raises:
            TimeoutElapsed: If the countdown expire.

        Args:
            limit (int): The countdown limit.
        '''

        self._task = self._loop.create_task(self._start_countdown(limit))

        try:
            await self._task
        except CancelledError:
            return

        raise TimeoutElapsed()

    async def _start_countdown(self, limit: int) -> None:
        '''A coroutine that await for n time (in seconds).

        Args:
            limit (int): The countdown limit.            
        '''

        await asyncio.sleep(limit)

    def stop(self) -> None:
        '''Cancel the self._start_countdown cotourine.
        '''

        if self._task and not self._task.done():
            self._task.cancel()


class TimeoutElapsed(Exception):
    '''The TimeoutElapsed Exception.

    Raised when the timeout had expired.
    '''




import pamqp
import aiormq
from aiormq import Channel, Connection

Props = aiormq.spec.Basic.Properties
DeliveredMessage = aiormq.types.DeliveredMessage
ArgumentsType = aiormq.types.ArgumentsType
GetEmpty, GetOk = aiormq.spec.Basic.GetEmpty, aiormq.spec.Basic.GetOk



'''
@desc    The queue class for manage queues.
@author  arthuchaut <arthuchaut@gmail.com>
@version 0.1.0
@date    2019-09-20
@note    0.1.0 (2019-09-20): Writed the first drafts.
'''


class Queue:
    '''The Queue class declaration.

    Attributes:
        queue (str): The queue name.
        passive (bool): Not create queue if it not exists.
        durable (bool): Not delete the queue after a broker restart.
        exclusive (bool): Delete the queue after closed the connection.
        auto_delete (bool): Delete queue when it empty from consumers.
        nowait (bool): Not send a reply callback.
        arguments (dict): Some arguments for the declaration.
    '''

    def __init__(
        self,
        queue: str = '',
        *,
        passive: bool = False,
        durable: bool = False,
        exclusive: bool = False,
        auto_delete: bool = False,
        nowait: bool = False,
        arguments: dict = None
    ):
        '''The Queue initializer.

        Args:
            queue (str): The queue name. Default to ''.
            passive (bool): Not create queue if it not exists.
                Default to False.
            durable (bool): Not delete the queue after a broker restart.
                Default to False.
            exclusive (bool): Delete the queue after closed the connection.
                Default to False.
            auto_delete (bool): Delete queue when it empty from consumers.
                Default to False.
            nowait (bool): Not send a reply callback.
                Default to False.
            arguments (dict): Some arguments for the declaration.
                Default to {}.
        '''

        self.queue: str = queue
        self.passive: bool = passive
        self.durable: bool = durable
        self.exclusive: bool = exclusive
        self.auto_delete: bool = auto_delete
        self.nowait: bool = nowait
        self.arguments: dict = arguments or {}



'''
@desc    The main class of the Symbios library.
@author  arthuchaut <arthuchaut@gmail.com>
@version 0.3.0
@date    2019-09-20
@note    0.1.0 (2019-09-20): Writed the first drafts.
@note    0.2.0 (2019-09-21): Implemented the class methods.
@note    0.3.0 (2019-09-22): Implemented some basic middlewares.
'''

from typing import Dict, Union, Callable, Awaitable, Any

from .utils import Props, ArgumentsType, Channel
from .connector import Connector
from .queue import Queue
from .exchange import Exchange
from .message import IncomingMessage, SendingMessage
from .middleware import MiddlewareLibrary, MiddlewareABC, Event
from .producer import Producer
from .consumer import Consumer
from .rpc import RPC

from middlewares.deserializer_middleware import DeserializerMiddleware
from middlewares.serializer_middleware import SerializerMiddleware
from .confirmation import EmitACK, ListenACK, ExchangeACK, QueueACK


class Symbios(Connector):
    '''The Symbios class declaration.

    Allows to communicate with the broker.

    Attributes:
        _midd_library (MiddlewareLibrary): The middleware library.
        rpc (RPC): The RPC instance.
    '''

    def __init__(self, **kwargs: Dict[str, Union[str, int]]):
        '''The Symbios initializer.

        Args:
            **kwargs (Dict[str, Union[str, int]]): The connection credentials.
                See the Connector __init__ documentation for more information.
        '''

        super().__init__(**kwargs)

        self._midd_library: MiddlewareLibrary = MiddlewareLibrary()
        self.rpc: RPC = RPC(self)
        self._init_standard_middlewares()

    def _init_standard_middlewares(self) -> None:
        '''Implement some basic middlewares.
        '''

        self.use(SerializerMiddleware(Event.ON_EMIT))
        self.use(DeserializerMiddleware(Event.ON_LISTEN))

    async def declare_queue(self, queue: Queue) -> QueueACK:
        '''Declare a queue to the broker.

        Args:
            queue (Queue): The queue to declare

        Returns:
            QueueACK: The validation of the declaration.
        '''

        chan: Channel = await self.channel

        confirmation = await chan.queue_declare(**queue.__dict__)

        return QueueACK(confirmation)

    async def declare_exchange(self, exchange: Exchange) -> ExchangeACK:
        '''Declare an exchange to the broker.

        Args:
            exchange (Exchange): The exchange to declare

        Returns:
            ExchangeACK: The validation of the declaration.
        '''

        chan: Channel = await self.channel

        confirmation = await chan.exchange_declare(**exchange.__dict__)

        return ExchangeACK(confirmation)

    async def emit(
        self,
        message: SendingMessage,
        *,
        exchange: Exchange = Exchange(),
        routing_key: str = None,
        props: Props = Props(),
        mandatory: bool = False,
        immediate: bool = False,
    ) -> None:
        '''Emit a message to the broker.

        Args:
            message (SendingMessage): The message to send.
            exchange (Exchange): the exchange name to bind with the 
                exchange type. Default to Exchange().
            routing_key (str): The message routing key. Default to None.
            props (Props): The message properties that contain the header.
                Default to an empty instance of Props.
            mandatory (bool): Tell if the message is important or not.
                Default to False.
            immediate (bool): Bypass the exchange type rules and send
                the message to the receiver directly if True.
                Default to False.
        '''

        producer: Producer = Producer(
            symbios=self,
            exchange=exchange,
            routing_key=routing_key,
            props=props,
            mandatory=mandatory,
            immediate=immediate,
            midd_library=self._midd_library,
        )

        return await producer.emit(message)

    async def listen(
        self,
        task: Callable[[object, IncomingMessage], None],
        *,
        queue: Queue = Queue(),
        no_ack: bool = False,
        exclusive: bool = False,
        arguments: ArgumentsType = None,
        consumer_tag: str = None,
    ) -> None:
        '''Listen a message from a broker queue.

        Args:
            task (Callable[[Symbios, IncomingMessage], None]): The
                task to call when a message arrives.
            queue (str): The queue to consume. Default to Queue().
            no_ack (bool): If deliver an aknowlegment or not. 
                Default to False.
            exclusive (bool): Only one consumer registered to 
                the targeted queue. Default to False.
            arguments (ArgumentsType): Some properties to the consumer.
                Default to None.
            consumer_tag (str): The consumer identity. Default to None.
        '''

        consumer: Consumer = Consumer(
            symbios=self,
            queue=queue,
            no_ack=no_ack,
            exclusive=exclusive,
            arguments=arguments,
            consumer_tag=consumer_tag,
            midd_library=self._midd_library,
        )

        return await consumer.listen(task)

    def use(self, midd: MiddlewareABC) -> None:
        '''Implement a new middleware for the consumer.

        All middlewares will be called before the main task.
        The IncomingMessage could be modified in.

        Args:
            midd (MiddlewareABC): The middleware instance to register to the
                middleware library.
        '''

        self._midd_library.append(midd)



'''
@desc    The exchange class for manage exchanges.
@author  arthuchaut <arthuchaut@gmail.com>
@version 0.1.0
@date    2019-09-20
@note    0.1.0 (2019-09-20): Writed the first drafts.
'''


class Exchange:
    '''The Exchange class declaration.

    Attributes:
        DIRECT (str): The default exhange type. Sends the message
            to alls the queues that are bound to the same routing_key that
            the message routing_key.

            Example: 
                If Producer::routing_key == Consumer::queue::binding_key
                then deliver the message to this queue.
        FANOUT (str): Sends the message to all the queues 
            that are bound to it. The routing_key will be ignored.

            Example:
                If Consumer::binding_key is bound to this exchange_type
                then deliver the message to this queue.
        TOPIC (str): Sends the message to all the queues that
            are bound to the same routing_key that the message routing_key
            and that the routing_key pattern of the bound queue is matching 
            with the message routing_key.

            Example:
                Producer::routing_key = 'my.routing.key'
                Consumer::binding_key = 'my.*.key'

                is_matching(Producer::routing_key, Consumer::binding_key)
                >>> True

                Producer::routing_key = 'his.routing.key'
                Consumer::binding_key = 'my.*.key'

                is_matching(Producer::routing_key, Consumer::binding_key)
                >>> False
        HEADERS (str): Sends the message to all queues at which 
            their routing_key matches to specific header attributes. 
            The routing_key is ignored.

            Example:
                Header::attr['x-match'] = 'my_value'
                Consumer::binding_key = 'my_value'

                Header::attr['x-match'] == Consumer::binding_key
                >>> True

                Header::attr['x-match'] = 'my_other_value'
                Consumer::binding_key = 'my_value'

                Header::attr['x-match'] == Consumer::binding_key
                >>> False
        exchange (str): The exchange name.
        exchange_type (str): The exchange type. See upper.
        passive (bool): Not create exchange if it not exists.
        durable (bool): Not delete the exchange after a broker restart.
        exclusive (bool): Delete the exchange after closed the connection.
        auto_delete (bool): Delete exchange when it empty from producers.
        nowait (bool): Not send a reply callback.
        arguments (dict): Some arguments for the declaration.
    '''

    DIRECT: str = 'direct'
    FANOUT: str = 'fanout'
    TOPIC: str = 'topic'
    HEADERS: str = 'headers'

    def __init__(
        self,
        exchange: str = '',
        *,
        exchange_type: str = DIRECT,
        passive: bool = False,
        durable: bool = False,
        auto_delete: bool = False,
        nowait: bool = False,
        arguments: dict = {}
    ):
        '''The Exchange initializer.

        Args:
            exchange (str): The exchange name. Default to ''.
            exchange_type (str): The exchange type. See upper.
                Default to DIRECT.
            passive (bool): Not create exchange if it not exists.
                Default to False.
            durable (bool): Not delete the exchange after a broker restart.
                Default to False.
            exclusive (bool): Delete the exchange after closed the connection.
                Default to False.
            auto_delete (bool): Delete exchange when it empty from producers.
                Default to False.
            nowait (bool): Not send a reply callback.
                Default to False.
            arguments (dict): Some arguments for the declaration.
                Default to {}.
        '''

        self.exchange: str = exchange
        self.exchange_type: str = exchange_type
        self.passive: bool = passive
        self.durable: bool = durable
        self.auto_delete: bool = auto_delete
        self.nowait: bool = nowait
        self.arguments: dict = arguments



'''
@desc    The consumer class for listen message from the broker.
@author  arthuchaut <arthuchaut@gmail.com>
@version 0.1.0
@date    2019-09-20
@note    0.1.0 (2019-09-20): Writed the first drafts.
'''

import asyncio

from typing import Callable, Awaitable, Any

from .utils import Channel, DeliveredMessage, ArgumentsType
from .message import IncomingMessage
from .queue import Queue
from .middleware import MiddlewareLibrary, Event
from .confirmation import ListenACK, QueueACK


class Consumer:
    '''The Consumer class declaration.

    Listen message from the broker.

    Attributes:
        symbios (Symbios): The Symbios instance.
        queue (str): The queue to consume.
        no_ack (bool): If deliver an aknowlegment or not.
        exclusive (bool): Only one consumer registered to the targeted queue.
        arguments (ArgumentsType): Some properties to the consumer.
        consumer_tag (str): The consumer identity.
        task (Callable[[Symbios, IncomingMessage], None]): 
            The task to call when a message arrives.
        declare_ok (Any): The future of the declared queue.
        consume_ok (Any): The future of the consumed queue.
        _midd_library (MiddlewareLibrary): The Symbios middleware library.
    '''

    def __init__(
        self,
        *,
        symbios: object,
        queue: Queue = Queue(),
        no_ack: bool = False,
        exclusive: bool = False,
        arguments: ArgumentsType = None,
        consumer_tag: str = None,
        midd_library: MiddlewareLibrary,
    ):
        '''The Consumer initializer.

        Args:
            symbios (Symbios): The Symbios instance.
            queue (str): The queue to consume. Default to Queue().
            no_ack (bool): If deliver an aknowlegment or not. 
                Default to False.
            exclusive (bool): Only one consumer registered to 
                the targeted queue. Default to False.
            arguments (ArgumentsType): Some properties to the consumer.
                Default to None.
            consumer_tag (str): The consumer identity. Default to None.
            midd_library (MiddlewareLibrary): The Symbios middleware library.
        '''

        self.symbios: object = symbios
        self.queue: Queue = queue
        self.no_ack: bool = no_ack
        self.exclusive: bool = exclusive
        self.arguments: ArgumentsType = arguments
        self.consumer_tag: str = consumer_tag
        self.task: Callable[[object, IncomingMessage], None] = None
        self.declare_ok: Any = None
        self.consume_ok: Any = None
        self._midd_library = midd_library

    async def listen(
        self, task: Callable[[object, IncomingMessage], None]
    ) -> ListenACK:
        '''Listen a queue.

        Args:
            task (Callable[[Symbios, IncomingMessage], None]): 
                The task to call when a message arrives.

        Returns:
            ListenACK: The consumer confirmation.
        '''

        self.task = task
        chann: Channel = await self.symbios.channel

        declare_ok = await chann.queue_declare(**self.queue.__dict__)
        self.declare_ok = QueueACK(declare_ok)

        consume_ok = await chann.basic_consume(
            self.declare_ok.confirmation.queue,
            self._embed,
            no_ack=self.no_ack,
            exclusive=self.exclusive,
            arguments=self.arguments,
            consumer_tag=self.consumer_tag,
        )

        self.consume_ok = ListenACK(consume_ok)

        return self.consume_ok

    async def _embed(self, message: DeliveredMessage) -> None:
        '''Embed the task with the Symbios parameters.
        
        Call the associated middlewares and then call the task.

        Args:
            message (DeliveredMessage): The aiormq message model.
        '''

        message: IncomingMessage = IncomingMessage(message)

        if not self._midd_library is None:
            await self._midd_library.run_until_end(
                self.symbios, message, Event.ON_LISTEN
            )

        await self.task(self.symbios, message)



'''
@desc    The body deserializer middleware for Symbios.
@author  arthuchaut <arthuchaut@gmail.com>
@version 0.1.0
@date    2019-09-22
@note    0.1.0 (2019-09-22): Writed the first drafts.
'''

from typing import Union, Dict, Any
import json

from symbios.middleware import MiddlewareABC
from symbios.message import IncomingMessage


class DeserializerMiddleware(MiddlewareABC):
    '''The DeserializerMiddleware class declaration.

    Attributes:
        _content_type_corr (Dict[str, Any]): The content-type
            correlation dictionnary.
    '''

    def __init__(self, event: int):
        '''The DeserializerMiddleware initializer.

        Args:
            event (int): The event associated with middleware.
        '''

        super().__init__(event)

        self._content_type_corr: Dict[str, Any] = {
            'text/plain': self._bytes_to_str,
            'application/json': self._bytes_to_dict,
        }

    async def execute(self, symbios: object, message: IncomingMessage) -> None:
        '''Implement to the message the deserialized format.
        '''

        content: Union[str, Dict[str, Any]] = None

        try:
            content = self._content_type_corr[message.props.content_type](
                message.body
            )
        except KeyError as e:
            content = self._bytes_to_str(message.body)

        setattr(message, 'deserialized', content)

    def _bytes_to_str(self, body: bytes) -> str:
        '''Parse the body to string.

        Args:
            body (bytes): The message body to parse.

        Returns:
            str: The body parsed.
        '''

        return body.decode()

    def _bytes_to_dict(self, body: bytes) -> Dict[str, Any]:
        '''Parse the body to Dict[str, Any].

        Args:
            body (bytes): The message body to parse.

        Returns:
            Dict[str, Any]: The body parsed.
        '''

        return json.loads(self._bytes_to_str(body))



'''
@desc    The message class.
@author  arthuchaut <arthuchaut@gmail.com>
@version 0.3.0
@date    2019-09-20
@note    0.1.0 (2019-09-20): Writed the first drafts.
@note    0.2.0 (2019-09-22): Redesigned message classes since 
                             the new middleware structure.
'''

from typing import Union, Dict, Any
import json

from .utils import DeliveredMessage, GetEmpty, GetOk, Props


class IncomingMessage:
    '''The IncomingMessage class declaration.

    This class will be instanciated when a message will incoming from
    the consumer.

    Attributes:
        delivery (Union[GetEmpty, GetOk]): The message ack tag.
        header (Any): The message header.
        props (Props): The message properties.
        body (bytes): The message content.
    '''

    def __init__(self, message: DeliveredMessage):
        '''The IncomingMessage initializer.

        Args:
            message (DeliveredMessage): The aiormq message object.
        '''

        self.delivery: Union[GetEmpty, GetOk] = message.delivery
        self.header: Any = message.header
        self.props: Props = message.header.properties
        self.body: bytes = message.body


class SendingMessage:
    '''The SendingMessage class declaration.

    Attributes:
        body (Union[str, Dict[str, Any]]): The message to send.
    '''

    def __init__(self, body: Union[str, Dict[str, Any]]):
        '''The SendingMessage initializer.

        Args:
            body (Union[str, Dict[str, Any]]): The message to send.
        '''

        self.body: Union[str, Dict[str, Any]] = body


class SendingMessageError(Exception):
    '''The SendingMessageError Exception class.
    '''

    def __init__(self, message: str):
        super().__init__(message)



from distutils.core import setup

setup(
    name='symbios',
    packages=['symbios', 'middlewares'],
    version='1.1.0',
    license='MIT',
    description='A simple asynchronous AMQP library writed for Python.',
    author='Arthuchaut',
    author_email='arthuchaut@gmail.com',
    url='https://github.com/Arthuchaut/symbios',
    download_url='https://github.com/Arthuchaut/symbios/archive/v1.1.0-dev.tar.gz',
    keywords=['Rabbitmq', 'Python', 'Async', 'Broker', 'AMQP'],
    install_requires=['aiormq'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
    ],
)



'''
@desc    The serializer middleware for Symbios.
@author  arthuchaut <arthuchaut@gmail.com>
@version 0.1.0
@date    2019-09-22
@note    0.1.0 (2019-09-22): Writed the first drafts.
'''

from typing import Union, Dict, Any
import json

from symbios.middleware import MiddlewareABC
from symbios.message import SendingMessage


class SerializerMiddleware(MiddlewareABC):
    '''The SerializerMiddleware class declaration.

    Attributes:
        _type_corr (Dict[str, Any]): The typing converter dictionnary.
    '''

    def __init__(self, event: int):
        '''The SerializerMiddleware initializer.

        Args:
            event (int): The event associated with middleware.
        '''

        super().__init__(event)

        self._type_corr: Dict[str, Any] = {
            str: self._str_to_bytes,
            int: self._str_to_bytes,
            float: self._str_to_bytes,
            dict: self._dict_to_bytes,
        }

    async def execute(self, symbios: object, message: SendingMessage) -> None:
        '''Serialize the message.

        Implement the serialized attribute to the message.

        Args:
            symbios (Symbios): The symbios instance.
            message (SendingMessage): The message to send. 

        Raises:
            SerializerMiddlewareError: If the type of message is 
                not serializable.
        '''

        try:
            setattr(
                message,
                'serialized',
                self._type_corr[type(message.body)](message.body),
            )
        except KeyError as e:
            raise SerializerMiddlewareError(
                f'Type {type(message.body).__name__} is not serializable.'
            )

    def _str_to_bytes(self, body: Union[str, int, float]) -> bytes:
        '''Parse the body string to bytes.

        Args:
            body (Union[str, int, float]): The message body to parse.

        Returns:
            bytes: The body parsed.
        '''

        return str(body).encode()

    def _dict_to_bytes(self, body: Dict[Any, Any]) -> bytes:
        '''Parse the Dict[Any, Any] body to a stringify JSON bytes.

        Args:
            body (Dict[Any, Any]): The message body to parse.

        Returns:
            bytes: The body parsed.
        '''

        return json.dumps(body).encode()


class SerializerMiddlewareError(Exception):
    '''The SerializerMiddlewareError Exception class.
    '''

    def __init__(self, message: str):
        super().__init__(message)
